"""
This linter checks for not used columns and tables by going through SELECT queries
"""
import logging

from collections import defaultdict, OrderedDict
from sql_metadata import get_query_columns, get_query_tables

from indexdigest.database import IndexDigestQueryError
from indexdigest.utils import LinterEntry, is_select_query


def get_used_tables_from_queries(database, queries):
    """
    :type database  indexdigest.database.Database
    :type queries list[str]
    :rtype: list[str]
    """
    logger = logging.getLogger(__name__)

    used_tables = []
    queries = filter(is_select_query, queries)

    for query in queries:
        # run EXPLAIN for each query from the log
        try:
            for row in database.explain_query(query):
                if row.get('table') is not None:
                    if row['table'] not in used_tables:
                        used_tables.append(row['table'])
                else:
                    # EXPLAIN may return "no matching row in const table"
                    logger.warning('EXPLAIN %s returned no table, falling back to SQL parsing',
                                   query)

                    # fall back to SQL query parsing
                    tables = get_query_tables(query)
                    if tables and tables[0] not in used_tables:
                        used_tables.append(tables[0])
        except IndexDigestQueryError:
            logger.error('Cannot explain the query: %s', query)

    return used_tables


def check_not_used_tables(database, queries):
    """
    :type database  indexdigest.database.Database
    :type queries list[str]
    :rtype: list[LinterEntry]
    """
    logger = logging.getLogger(__name__)

    # get database meta-data
    tables = database.get_tables()

    # analyze only SELECT queries from the log
    used_tables = get_used_tables_from_queries(database, queries)
    logger.info("These tables were used by provided queries: %s", used_tables)

    # now check which tables were not used
    not_used_tables = [table for table in tables if table not in used_tables]

    # generate reports
    for table in not_used_tables:
        metadata = database.get_table_metadata(table)
        context = OrderedDict()

        context['schema'] = database.get_table_schema(table)
        context['table_size_mb'] = \
            1. * (metadata['data_size'] + metadata['index_size']) / 1024 / 1024
        context['rows_estimated'] = database.get_table_rows_estimate(table)

        yield LinterEntry(linter_type='not_used_tables', table_name=table,
                          message='"{}" table was not used by provided queries'.format(table),
                          context=context)


def check_not_used_columns(database, queries):
    """
    :type database  indexdigest.database.Database
    :type queries list[str]
    :rtype: list[LinterEntry]
    :raises Exception
    """
    logger = logging.getLogger(__name__)

    # analyze only SELECT queries from the log
    queries = list(filter(is_select_query, queries))

    used_tables = get_used_tables_from_queries(database, queries)
    used_columns = defaultdict(list)

    logger.info("Will check these tables: %s", used_tables)

    # analyze given queries and collect used columns for each table
    for query in queries:
        tables = get_query_tables(query)
        if tables:
            columns = get_query_columns(query)

            # print(query, table, columns)

            # add used columns per table
            # FIXME: assume we're querying just a single table for now
            used_columns[tables[0]] += columns
        else:
            logger.error('Unable to extract tables and columns used from the query: %s',
                         query)

    # analyze table schemas and report not used columns for each table
    for table in used_tables:
        logger.info("Checking %s table", table)
        table_columns = database.get_table_columns(table)

        # now get the difference and report them
        not_used_columns = [
            column for column in table_columns
            if column.name not in set(used_columns[table])
        ] if table_columns else []

        for column in not_used_columns:
            yield LinterEntry(linter_type='not_used_columns', table_name=table,
                              message='"{}" column was not used by provided queries'.format(column),
                              context={'column_name': column.name, 'column_type': column.type})



"""
This linter reports text columns that have characters encoding set to latin1
"""
from collections import OrderedDict

from indexdigest.utils import LinterEntry


def check_latin_columns(database):
    """
    :type database  indexdigest.database.Database
    :rtype: list[LinterEntry]
    """
    for table in database.get_tables():
        for column in database.get_table_columns(table):
            if not column.is_text_type():
                continue

            # ignore utf8 columns
            if column.character_set in ['utf8']:
                continue

            # print([table, column, column.character_set, column.collation])

            context = OrderedDict()
            context['column'] = column.name
            context['column_character_set'] = column.character_set
            context['column_collation'] = column.collation

            yield LinterEntry(linter_type='non_utf_columns', table_name=table,
                              message='"{}" text column has "{}" character set defined'.
                              format(column.name, column.character_set),
                              context=context)



"""
This linter reports tables with just a single column
"""
from indexdigest.utils import LinterEntry


def check_single_column(database):
    """
    :type database  indexdigest.database.Database
    :rtype: list[LinterEntry]
    """
    tables = [
        table
        for table in database.get_tables()
        if len(database.get_table_columns(table)) == 1
    ]

    for table in tables:
        yield LinterEntry(linter_type='single_column', table_name=table,
                          message='"{}" has just a single column'.
                          format(table),
                          context={'schema': database.get_table_schema(table)})



# -*- coding: utf-8 -*-
"""
Provides --format=plain results formatter
"""
from termcolor import colored

import indexdigest
from indexdigest.utils import LinterEntry


def format_context(context):
    """
    :type context dict
    :rtype: str
    """
    return '\n  '.join([
        "- {key}: {value}".format(
            key=colored(key, color='green', attrs=['bold']),
            value=str(value).replace("\n", "\n    ")
        )
        for (key, value) in context.items()
    ])


def format_plain(database, reports):
    """
    :type database indexdigest.database.Database
    :type reports list
    :rtype: str
    """
    out = ''

    # cast to a list (to be able to count reports)
    reports = list(reports)

    # emit results
    line = '-' * 60 + "\n"

    out += line
    out += 'Found {} issue(s) to report for "{}" database\n'.format(
        len(reports), database.db_name)
    out += line
    out += 'MySQL v{} at {}\n'.format(
        database.get_server_version(), database.get_server_hostname())
    out += 'index-digest v{}\n'.format(indexdigest.VERSION)
    out += line

    if reports:
        for report in reports:
            assert isinstance(report, LinterEntry)

            out += colored(report.linter_type, color='blue', attrs=['bold']) + \
                ' → table affected: ' + \
                colored(report.table_name, attrs=['bold']) + \
                '\n'

            out += colored(
                '\n{} {}\n'.format(colored('✗', color='red', attrs=['bold']), report.message),
                color='white')

            if report.context is not None:
                out += '\n  {}\n'.format(format_context(report.context))

            out += '\n'
            out += line

        out += 'Queries performed: {}'.format(len(database.get_queries()))
        # out += '\n'.join(map(str, database.get_queries())))
    else:
        out += 'Jolly, good! No issues to report'

    return out



"""
This linter reports missing primary / unique index
"""
from collections import OrderedDict

from indexdigest.utils import LinterEntry


def check_missing_primary_index(database):
    """
    :type database  indexdigest.database.Database
    :rtype: list[LinterEntry]
    """
    for table in database.get_tables():
        # list non-primary (and non-unique) indices only
        # @see https://bugs.mysql.com/bug.php?id=76252
        # @see https://github.com/Wikia/app/pull/9863
        indices = [
            index for index in database.get_table_indices(table)
            if index.is_primary or index.is_unique
        ]

        if indices:
            # so we have at least one primary or unique index defined
            continue

        context = OrderedDict()
        context['schema'] = database.get_table_schema(table)

        yield LinterEntry(linter_type='missing_primary_index', table_name=table,
                          message='"{}" table does not have any primary or unique index'.
                          format(table),
                          context=context)



"""
This module contains utility functions and classes
"""
try:
    from urlparse import urlparse  # Python2
except ImportError:
    from urllib.parse import urlparse  # Python3

import functools
import logging


def parse_dsn(dsn):
    """
    Parser given Data Source Name string into an object that can be passed to a database connector.

    Example: mysql://alex:pwd@localhost/test
    DSN: <scheme>://<netloc>/<path>;<params>?<query>#<fragment>

    @see https://mysqlclient.readthedocs.io/user_guide.html#mysqldb

    :type dsn str
    :rtype: dict
    """
    parsed = urlparse(dsn)

    return {
        'host': parsed.hostname,
        'user': parsed.username,
        'passwd': parsed.password,
        'db': str(parsed.path).lstrip('/')
    }


def is_select_query(query):
    """
    :type query str
    :rtype bool
    """
    return query.lstrip().lower().startswith('select')


def explain_queries(database, queries):
    """
    Yields EXPLAIN result rows for given queries

    :type database  indexdigest.database.Database
    :type queries list[str]
    :rtype: tuple[str,str,str,str]
    """
    # analyze only SELECT queries from the log
    for query in filter(is_select_query, queries):
        try:
            for row in database.explain_query(query):
                table_used = row['table']
                index_used = row['key']

                yield (query, table_used, index_used, row)
        except IndexDigestError:
            logger = logging.getLogger('explain_queries')
            logger.error('Cannot explain the query: %s', query)


def shorten_query(query, max_len=50):
    """
    :type query str
    :type max_len int
    :rtype: str
    """
    query = query.rstrip('; ')
    return '{}...'.format(query[:max_len]) if len(query) > max_len else query


def memoize(func):
    """
    Memoization pattern implemented

    :type func
    :rtype func
    """
    # @see https://medium.com/@nkhaja/memoization-and-decorators-with-python-32f607439f84
    cache = func.cache = {}

    @functools.wraps(func)
    def memoized_func(*args, **kwargs):
        """
        :type args
        :type kwargs
        """
        key = str(args) + str(kwargs)
        if key not in cache:
            cache[key] = func(*args, **kwargs)
        return cache[key]
    return memoized_func


class LinterEntry(object):
    """
    Wraps a single linter entry. Various formatters may display this data differently.
    """
    def __init__(self, linter_type, table_name, message, context=None):
        """
        :type linter_type str
        :type table_name str
        :type message str
        :type context dict
        """
        self.linter_type = linter_type
        self.table_name = table_name
        self.message = message
        self.context = context

    def __str__(self):
        return '{table_name}: {message}'.format(
            table_name=self.table_name, message=self.message)


class IndexDigestError(Exception):
    """
    index-digest base exception class
    """
    pass



"""
This linter reports tables with "test" word in their name
"""
import re

from indexdigest.utils import LinterEntry


def is_test_table(table_name):
    """
    :type table_name str
    :rtype: bool
    """
    return re.search(r'(^|_)test(_|$)', table_name) is not None


def check_test_tables(database):
    """
    :type database  indexdigest.database.Database
    :rtype: list[LinterEntry]
    """
    test_tables = [
        table for table in database.get_tables()
        if is_test_table(table)
    ]

    for table in test_tables:
        yield LinterEntry(linter_type='test_tables', table_name=table,
                          message='"{}" seems to be a test table'.
                          format(table),
                          context={'schema': database.get_table_schema(table)})



"""
This linter checks INSERT IGNORE queries

If you use the IGNORE modifier, errors that occur while executing the INSERT statement are ignored.
For example, without IGNORE, a row that duplicates an existing UNIQUE index or PRIMARY KEY value
in the table causes a duplicate-key error and the statement is aborted. With IGNORE, the row is
discarded and no error occurs. Ignored errors generate warnings instead.

Data conversions that would trigger errors abort the statement if IGNORE is not specified.
With IGNORE, invalid values are adjusted to the closest values and inserted; warnings
are produced but the statement does not abort.

@see https://medium.com/legacy-systems-diary/things-to-avoid-episode-1-insert-ignore-535b4c24406b
"""
import re

from collections import OrderedDict
from sql_metadata import get_query_tables

from indexdigest.utils import LinterEntry, shorten_query


def remove_comments(sql):
    """
    :type sql str
    :rtype: str
    """
    return re.sub(r'/\*[^*]+\*/', '', sql)


def is_insert_ignore_query(sql):
    """
    :type sql str
    :rtype: bool
    """
    sql = remove_comments(sql).lstrip()
    return re.match(r'^INSERT\s+IGNORE\s', sql, flags=re.IGNORECASE) is not None


def check_insert_ignore_queries(database, queries):
    """
    :type database  indexdigest.database.Database
    :type queries list[str]
    :rtype: list[LinterEntry]
    """
    queries = [query for query in queries if is_insert_ignore_query(query)]

    for query in queries:
        table_used = get_query_tables(query)[0]

        context = OrderedDict()
        context['query'] = query
        context['schema'] = database.get_table_schema(table_used)

        yield LinterEntry(linter_type='insert_ignore', table_name=table_used,
                          message='"{}" query uses a risky INSERT IGNORE'.
                          format(shorten_query(query)),
                          context=context)



"""
This linter checks for select queries with HAVING clause
"""
from sqlparse.tokens import Keyword
from sql_metadata import preprocess_query, get_query_tables, get_query_tokens

from indexdigest.utils import LinterEntry, shorten_query, is_select_query


def query_has_having_clause(query):
    """
    Checks if provided query uses HAVING clause
    :type query str
    :rtype bool
    """
    if not is_select_query(query):
        return False

    query = preprocess_query(query)
    tokens = get_query_tokens(query)

    for token in tokens:
        if token.ttype is Keyword and str(token).upper() == 'HAVING':
            return True

    return False


def check_having_clause(_, queries):
    """
    :type queries list[str]
    :rtype: list[LinterEntry]
    """
    queries_with_having_clause = [
        query for query in queries
        if query_has_having_clause(query)
    ]

    for query in queries_with_having_clause:
        table_name = get_query_tables(query)[0]

        yield LinterEntry(linter_type='having_clause', table_name=table_name,
                          message='"{}" query uses HAVING clause'.
                          format(shorten_query(query)),
                          context={"query": query})



"""
Provides --format=syslog results formatter - pushes JSON messages via syslog
"""
from __future__ import absolute_import

import json
import syslog

from collections import OrderedDict

import indexdigest


def format_report(database, report):
    """
    :type database indexdigest.database.Database
    :type report indexdigest.utils.LinterEntry
    :rtype: str
    """
    res = OrderedDict()

    res['appname'] = 'index-digest'

    res['meta'] = OrderedDict()
    res['meta']['version'] = 'index-digest v{}'.format(indexdigest.VERSION)
    res['meta']['database_name'] = database.db_name
    res['meta']['database_host'] = database.get_server_hostname()
    res['meta']['database_version'] = 'MySQL v{}'.format(database.get_server_version())

    res['report'] = OrderedDict()
    res['report']['type'] = report.linter_type
    res['report']['table'] = report.table_name
    res['report']['message'] = report.message

    if report.context:
        res['report']['context'] = report.context

    return json.dumps(res)


def format_syslog(database, reports, ident='index-digest'):
    """
    :type database indexdigest.database.Database
    :type reports list
    :type ident str
    :rtype: str
    """
    syslog.openlog(ident=ident, logoption=syslog.LOG_PID, facility=syslog.LOG_USER)

    for report in reports:
        syslog.syslog(format_report(database, report))

    syslog.closelog()
    return ''



"""index_digest

Analyses your database queries and schema and suggests indices improvements.

Usage:
  index_digest DSN [--sql-log=<file>] [--format=<formatter>]
  index_digest (-h | --help)
  index_digest --version

Options:
  DSN               Data Source Name of database to check
  --sql-log=<file>  Text file with SQL queries to check against the database
  --format=<formatter>  Use a given results formatter (plain, syslog, yaml)
  -h --help         Show this screen.
  --version         Show version.

Examples:
  index_digest mysql://username:password@localhost/dbname
  index_digest mysql://index_digest:qwerty@localhost/index_digest --sql-log=sql.log

Visit <https://github.com/macbre/index-digest>
"""
from __future__ import print_function

import logging
from itertools import chain
from os import getenv

from docopt import docopt

import indexdigest
from indexdigest.database import Database
from indexdigest.utils import IndexDigestError
from indexdigest.formatters import \
    format_plain, \
    format_syslog, \
    format_yaml
from indexdigest.linters import \
    check_queries_using_filesort, check_queries_using_temporary, \
    check_not_used_indices, check_queries_not_using_indices, \
    check_not_used_tables, check_not_used_columns, \
    check_redundant_indices, \
    check_full_table_scan, \
    check_latin_columns, \
    check_selects_with_like, \
    check_missing_primary_index, \
    check_test_tables, \
    check_insert_ignore_queries, \
    check_single_column, \
    check_empty_tables, \
    check_select_star, \
    check_having_clause


def main():
    """ Main entry point for CLI"""
    logger = logging.getLogger(__name__)

    arguments = docopt(__doc__, version='index_digest {}'.format(indexdigest.VERSION))
    logger.debug('Options: %s', arguments)

    if 'DSN' not in arguments:
        return

    # connect to the database
    database = Database.connect_dsn(arguments['DSN'])
    logger.debug('Connected to MySQL server v%s', database.get_server_version())

    # read SQL log file (if provided)
    sql_log = arguments.get('--sql-log')
    if sql_log:
        logger.debug('Trying to open SQL log file: %s', sql_log)

        with open(sql_log) as log_file:
            queries = log_file.readlines()
            queries = list(map(str.strip, queries))  # remove trailing spaces
            logger.debug('Got %d entries in SQL log file', len(queries))
    else:
        queries = None

    # run all checks
    reports = chain(
        check_redundant_indices(database),
        check_latin_columns(database),
        check_missing_primary_index(database),
        check_test_tables(database),
        check_single_column(database),
        check_empty_tables(database),
    )

    # checks that use SQL log
    if queries:
        reports = chain(
            reports,
            check_not_used_indices(database, queries=queries),
            check_not_used_tables(database, queries=queries),
            check_not_used_columns(database, queries=queries),
            check_queries_not_using_indices(database, queries=queries),
            check_queries_using_filesort(database, queries=queries),
            check_queries_using_temporary(database, queries=queries),
            check_full_table_scan(database, queries=queries),
            check_selects_with_like(database, queries=queries),
            check_insert_ignore_queries(database, queries=queries),
            check_select_star(database, queries=queries),
            check_having_clause(database, queries=queries),
        )

    # handle --format
    formatter = arguments.get('--format') or 'plain'
    logger.info("Using formatter: %s", formatter)

    if formatter == 'plain':
        print(format_plain(database, reports))
    elif formatter == 'syslog':
        ident = getenv('SYSLOG_IDENT', 'index-digest')
        logger.info('Using syslog ident: %s', ident)
        print(format_syslog(database, reports, ident))
    elif formatter == 'yaml':
        print(format_yaml(database, reports))
    else:
        raise IndexDigestError('Unknown formatter provided: {}'.format(formatter))



"""
This linter checks for redundant indices from a given set of them
"""
import logging

from collections import OrderedDict

from indexdigest.utils import LinterEntry


def get_redundant_indices(indices):
    """
    :type indices list[indexdigest.schema.Index]
    :rtype: list[tuple]
    """
    redundant_indices = []

    for index in indices:
        for compare in indices:
            if index.is_covered_by(compare):
                redundant_indices.append((index, compare, ))

    return redundant_indices


def check_redundant_indices(database):
    """
    :type database  indexdigest.database.Database
    :rtype: list[LinterEntry]
    """
    logger = logging.getLogger(__name__)

    for table in database.get_tables():
        logger.info("Checking %s table", table)

        indices = database.get_table_indices(table)
        meta = database.get_table_metadata(table)
        schema = database.get_table_schema(table)

        redundant_indices = set()

        for (redundant_index, suggested_index) in get_redundant_indices(indices):
            # the index we're about to suggest was reported as redundant - #48
            if suggested_index in redundant_indices:
                continue

            context = OrderedDict()
            context['redundant'] = str(redundant_index)
            context['covered_by'] = str(suggested_index)
            context['schema'] = schema
            context['table_data_size_mb'] = 1. * meta['data_size'] / 1024 / 1024
            context['table_index_size_mb'] = 1. * meta['index_size'] / 1024 / 1024

            # add to the list to avoid redundant indices being reported in a loop - #48
            redundant_indices.add(redundant_index)

            yield LinterEntry(linter_type='redundant_indices', table_name=table,
                              message='"{}" index can be removed as redundant (covered by "{}")'.
                              format(redundant_index.name, suggested_index.name),
                              context=context)



"""
A helper script used to create files for new linter
"""
import logging
import re
import sys

logging.basicConfig(
    level=logging.DEBUG,
    format='%(levelname)-8s %(message)s',
)


def add_linter(linter_id, linter_name):
    """
    :type linter_id int
    :type linter_name str
    """
    logger = logging.getLogger('add_linter')

    # normalize values
    linter_id_fmt = '{:04d}'.format(linter_id)
    linter_name = re.sub(r'[^a-z]+', '-', linter_name.strip().lower())

    logger.info("Creating a new linter: %s - %s ...", linter_id_fmt, linter_name)

    # /sql directory
    sql_name = 'sql/{}-{}'.format(linter_id_fmt, linter_name.replace('_', '-'))
    logger.info("Add SQL schema and log files (%s) ...", sql_name)

    with open(sql_name + '.sql', 'wt') as file_name:
        # 0002_not_used_indices
        table_name = '{}_{}'.format(linter_id_fmt, linter_name.replace('-', '_'))

        file_name.writelines([
            '-- Report ...\n',
            '--\n',
            '-- https://github.com/macbre/index-digest/issues/{}\n'.format(linter_id),
            'DROP TABLE IF EXISTS `{}`;\n'.format(table_name),
            'CREATE TABLE `{}` (\n'.format(table_name),
            '-- \n',
            ');\n',
        ])

        logger.info('... %s created', file_name.name)

    with open(sql_name + '-log', 'wt') as file_name:
        file_name.writelines([
            '-- \n',
        ])

        logger.info('... %s created', file_name.name)

    # /indexdigest/linters directory
    linter_name = linter_name.replace('-', '_')
    logger.info("Add a Python code for %s linter ...", linter_name)

    with open('indexdigest/linters/linter_{}_{}.py'.
              format(linter_id_fmt, linter_name), 'wt') as file_name:
        file_name.writelines([
            '"""\n',
            'This linter checks for ...\n',
            '"""\n',
            'from collections import defaultdict\n',
            '\n',
            'from indexdigest.utils import LinterEntry, explain_queries\n',
            '\n',
            '\n',
            'def check_{}(database, queries):\n'.format(linter_name),
            '    """\n',
            '    :type database  indexdigest.database.Database\n',
            '    :type queries list[str]\n',
            '    :rtype: list[LinterEntry]\n',
            '    """\n',
            '    yield LinterEntry(linter_type=\'{}\', table_name=table_name,\n'.
            format(linter_name),
            '                      message=\'"{}" ...\'.\n',
            '                      format("foo"),\n',
            '                      context={"foo": str("bar")})\n',
        ])

        logger.info('... %s created', file_name.name)

    logger.info("Add a test ...")

    with open('indexdigest/test/linters/test_{}_{}.py'.format(linter_id_fmt, linter_name), 'wt') \
            as file_name:
        file_name.writelines([
            'from __future__ import print_function\n',
            '\n',
            'from unittest import TestCase\n',
            '\n',
            'from indexdigest.linters.linter_{0}_{1} import check_{1}\n'.
            format(linter_id_fmt, linter_name),
            'from indexdigest.test import DatabaseTestMixin, read_queries_from_log\n',
            '\n',
            '\n',
            'class TestLinter(TestCase, DatabaseTestMixin):\n',
            '\n',
            '    def test_{}(self):\n'.format(linter_name),
            '        pass\n',
        ])

        logger.info('... %s created', file_name.name)


def main():
    """
    usage: add_linter 89 empty_tables
    """
    linter_id = int(sys.argv[1])
    linter_name = str(sys.argv[2])

    add_linter(linter_id, linter_name)



"""
This linter checks for not used indices by going through SELECT queries
"""
import logging

from collections import defaultdict

from indexdigest.utils import LinterEntry, explain_queries


def check_not_used_indices(database, queries):
    """
    :type database  indexdigest.database.Database
    :type queries list[str]
    :rtype: list[LinterEntry]
    """
    logger = logging.getLogger(__name__)

    used_indices = defaultdict(list)

    # EXPLAIN each query
    for (query, table_used, index_used, _) in explain_queries(database, queries):
        if index_used is not None:
            logger.info("Query <%s> uses %s index on `%s` table", query, index_used, table_used)
            used_indices[table_used].append(index_used)

    # analyze all tables used by the above queries
    # print(used_indices)
    for table_name in used_indices.keys():
        for index in database.get_table_indices(table_name):

            if index.name not in used_indices[table_name]:
                yield LinterEntry(linter_type='not_used_indices', table_name=table_name,
                                  message='"{}" index was not used by provided queries'.
                                  format(index.name),
                                  context={"not_used_index": str(index)})



"""
This linter checks SELECT queries that do not use indices
"""
from collections import OrderedDict

from indexdigest.utils import LinterEntry, explain_queries, shorten_query


def check_queries_not_using_indices(database, queries):
    """
    :type database  indexdigest.database.Database
    :type queries list[str]
    :rtype: list[LinterEntry]
    """
    for (query, table_used, index_used, explain_row) in explain_queries(database, queries):
        # print(query, explain_row)

        # EXPLAIN can return no matching row in const table in Extra column.
        # Do not consider this query as not using an index. -- see #44
        if explain_row['Extra'] in [
                'Impossible WHERE noticed after reading const tables',
                'no matching row in const table',
        ]:
            continue

        if index_used is None:
            context = OrderedDict()
            context['query'] = query

            # https://dev.mysql.com/doc/refman/5.7/en/explain-output.html#explain-extra-information
            context['explain_extra'] = explain_row['Extra']
            context['explain_rows'] = explain_row['rows']
            context['explain_filtered'] = explain_row.get('filtered')  # can be not set
            context['explain_possible_keys'] = explain_row['possible_keys']

            yield LinterEntry(linter_type='queries_not_using_index', table_name=table_used,
                              message='"{}" query did not make use of any index'.
                              format(shorten_query(query)),
                              context=context)



"""
This linter checks for select queries with * wildcard
"""
from sql_metadata import preprocess_query, get_query_tables, get_query_tokens, Wildcard

from indexdigest.utils import LinterEntry, shorten_query, is_select_query


def is_wildcard_query(query):
    """
    Checks if provided query selects using a * wildcard
    :type query str
    :rtype bool
    """
    if not is_select_query(query):
        return False

    query = preprocess_query(query)
    tokens = get_query_tokens(query)

    last_token = None

    for token in tokens:
        if token.ttype is Wildcard:
            # print([query, token, 'last token', last_token])

            # check what was before the wildcard
            # count(*) ?
            if last_token and str(last_token) not in ['(']:
                return True
        else:
            last_token = token

    return False


def check_select_star(_, queries):
    """
    :type queries list[str]
    :rtype: list[LinterEntry]
    """
    queries_with_wildcard = [
        query for query in queries
        if is_wildcard_query(query)
    ]

    for query in queries_with_wildcard:
        table_name = get_query_tables(query)[0]

        yield LinterEntry(linter_type='select_star', table_name=table_name,
                          message='"{}" query uses SELECT *'.
                          format(shorten_query(query)),
                          context={"query": query})



"""
Contains linters used to check the database for improvements.
"""
# expose linters
from .linter_0002_not_used_indices import check_not_used_indices
from .linter_0004_redundant_indices import check_redundant_indices
from .linter_0006_not_used_columns_and_tables import check_not_used_tables, check_not_used_columns
from .linter_0019_queries_not_using_indices import check_queries_not_using_indices
from .linter_0020_filesort_temporary_table import \
    check_queries_using_filesort, check_queries_using_temporary
from .linter_0026_full_table_scan import check_full_table_scan
from .linter_0027_selects_with_like import check_selects_with_like
from .linter_0032_utf_latin_columns import check_latin_columns
from .linter_0034_missing_primary_index import check_missing_primary_index
from .linter_0070_insert_ignore import check_insert_ignore_queries
from .linter_0074_single_column import check_single_column
from .linter_0075_test_tables import check_test_tables
from .linter_0089_empty_tables import check_empty_tables
from .linter_0092_select_star import check_select_star
from .linter_0093_having_clause import check_having_clause



"""
Provides --format=yaml results formatter
"""
from __future__ import absolute_import

from collections import OrderedDict

import yaml
import yamlordereddictloader

import indexdigest


def format_report(report):
    """
    :type report indexdigest.utils.LinterEntry
    :rtype: OrderedDict
    """
    res = OrderedDict()

    res['type'] = report.linter_type
    res['table'] = report.table_name
    res['message'] = report.message

    if report.context:
        res['context'] = report.context

    return res


def format_yaml(database, reports):
    """
    :type database indexdigest.database.Database
    :type reports list
    :rtype: str
    """
    report = OrderedDict()

    report['meta'] = OrderedDict()
    report['meta']['version'] = 'index-digest v{}'.format(indexdigest.VERSION)
    report['meta']['database_name'] = database.db_name
    report['meta']['database_host'] = database.get_server_hostname()
    report['meta']['database_version'] = 'MySQL v{}'.format(database.get_server_version())

    report['reports'] = [format_report(item) for item in reports]

    return yaml.dump(report,
                     Dumper=yamlordereddictloader.Dumper,
                     default_flow_style=False,
                     explicit_start=True,
                     explicit_end=True)



"""
This linter checks SELECT queries that use LIKE '%foo' conditions
"""
import re

from collections import OrderedDict

from indexdigest.utils import LinterEntry, explain_queries, shorten_query


def query_uses_leftmost_like(query):
    """
    Returns True for queries with LIKE '%foo' conditions

    :type query str
    :rtype: bool
    """
    # quit fast
    if 'like' not in query.lower():
        return False

    matches = re.search(r'LIKE\s\s?[\'"]%\w', query, flags=re.IGNORECASE)
    return matches is not None


def check_selects_with_like(database, queries):
    """
    :type database  indexdigest.database.Database
    :type queries list[str]
    :rtype: list[LinterEntry]
    """
    for (query, table_used, index_used, explain_row) in explain_queries(database, queries):
        if index_used is None and query_uses_leftmost_like(query):
            context = OrderedDict()
            context['query'] = query

            # https://dev.mysql.com/doc/refman/5.7/en/explain-output.html#explain-extra-information
            context['explain_extra'] = explain_row['Extra']
            context['explain_rows'] = explain_row['rows']

            yield LinterEntry(linter_type='selects_with_like', table_name=table_used,
                              message='"{}" query uses LIKE with left-most wildcard'.
                              format(shorten_query(query)),
                              context=context)



"""
This linter checks for empty tables
"""
from indexdigest.utils import LinterEntry


def check_empty_tables(database):
    """
    :type database  indexdigest.database.Database
    :rtype: list[LinterEntry]
    """
    empty_tables = [
        table for table in database.get_tables()
        # use both "information_schema" and "explain select count(*)" based methods
        # to get the rows count estimate
        if database.get_table_metadata(table).get('rows') == 0
        or database.get_table_rows_estimate(table) == 0
    ]

    for table in empty_tables:
        yield LinterEntry(linter_type='empty_tables', table_name=table,
                          message='"{}" table has no rows, is it really needed?'.format(table),
                          context={'schema': database.get_table_schema(table)})



from setuptools import setup, find_packages

from indexdigest import VERSION

# @see https://github.com/pypa/sampleproject/blob/master/setup.py
setup(
    name='indexdigest',
    version=VERSION,
    author='Maciej Brencz',
    author_email='maciej.brencz@gmail.com',
    license='MIT',
    description='Analyses your database queries and schema and suggests indices and schema improvements',
    url='https://github.com/macbre/index-digest',
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Topic :: Database',

        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    packages=find_packages(),
    install_requires=[
        'docopt==0.6.2',
        'coverage==4.4.2',
        'pylint==1.8.1',
        'pytest==3.2.3',
        'PyYAML==3.12',
        'mysqlclient==1.3.12',
        'sql_metadata==1.0.2',
        'termcolor==1.1.0',
        'yamlordereddictloader==0.4.0'
    ],
    entry_points={
        'console_scripts': [
            'add_linter=indexdigest.cli.add_linter:main',  # creates a new linter from a template
            'index_digest=indexdigest.cli.script:main',
        ],
    }
)
